/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <malloc.h>
#include <string.h>

#include "glass.h"

/* Saves an object to a file */
int _glassSaveObject(GlassObject *object, const char *fname, const char *comment)
{
  FILE *output_file;

  output_file = fopen(fname, "w");
  if(output_file == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to open file %s for saving.\n", fname);
    return FALSE;
  }

  /* Write a comment */
  if(comment == NULL)
    fprintf(output_file, "# A GLASS object, see glass.sourceforge.net for details\n");
  else
    fprintf(output_file, "#%s\n", comment);

  /* Write the name of the object, and the glass version */
  fprintf(output_file, "name\t\"%s\"\n", object->name);
  fprintf(output_file, "version\t\"%s\"\n", glassGetVersion());
  fprintf(output_file, "\n");

  /* Write the textures */
  _glassWriteTextures(output_file, fname, object);
  /* Write the materials */
  _glassWriteMaterials(output_file, object);
  /* Write the triangle lists */
  _glassWriteTriangleLists(output_file, object);
  /* Write the variables */
  _glassWriteVariables(output_file, object);
  /* Write the active points */
  _glassWriteActivePoints(output_file, object);
  /* Write the components */
  _glassWriteComponents(output_file, object);

  fclose(output_file);

  return TRUE;
}

/* Write all the textures into the file */
int _glassWriteTextures(FILE *output_file, const char *fname, GlassObject *object)
{
  GlassTexture *mat_ptr = object->textures;

  while(mat_ptr != NULL)
  {
    _glassWriteTexture(output_file, fname, mat_ptr);
    fprintf(output_file, "\n");

    mat_ptr = mat_ptr->next;
  }

  return TRUE;
}

/* Write a texture to the file */
int _glassWriteTexture(FILE *output_file, const char *fname, GlassTexture *texture)
{
  char *full_object_fname, *full_image_fname;
  char *object_fname, *image_fname, *t;
  char final_image_fname[300];
  int i, last_slash, max, ndirs;

  /* Begin a texture */
  fprintf(output_file, "begin texture\n");

  fprintf(output_file, "  name\t\"%s\"\n", _glassTextureGetName(texture));
  if(texture->image_file)
  {
    /* Change the absolute filename to a relative one */

    /* Copy the file names */
    full_object_fname = strdup(fname);
    full_image_fname = strdup(texture->image_file);
    object_fname = full_object_fname;
    image_fname = full_image_fname;
    /* Remove the root '/'s */
    if(*object_fname == '/')
      object_fname++;
    if(*image_fname == '/')
      image_fname++;

    /* Remove the file from the object fname */
    for(t = &object_fname[strlen(object_fname)]; t != object_fname && *t != '/'; t--);
    *(t+1) = '\0';
  
    /* Remove as many shared dirs as possible */
    last_slash = -1;
    max = strlen(object_fname) > strlen(image_fname) ?
      strlen(image_fname) : strlen(object_fname);
    for(i = 0; i < max && object_fname != '\0' && (object_fname[i] == image_fname[i]); i++)
    {
      if(object_fname[i] == '/' && image_fname[i] == '/')
	last_slash = i;
    }
    object_fname = &object_fname[last_slash+1];
    image_fname = &image_fname[last_slash+1];
  
    /* Add as many ..'s are required */
    ndirs = 0;
    for(i = 0; i < strlen(object_fname); i++)
      if(object_fname[i] == '/')
	ndirs++;

    final_image_fname[0] = '\0';
    for(i = 0; i < ndirs; i++)
      strncat(final_image_fname, "../", 300);
  
    /* Add the filename and its unique dirs to the end */
    strncat(final_image_fname, image_fname, 300);
  
    free(full_object_fname);
    free(full_image_fname);

    fprintf(output_file, "  image\t\"%s\"\n", final_image_fname);
  }
  fprintf(output_file, "end\n");

  return TRUE;
}

/* Write all the materials into the file */
int _glassWriteMaterials(FILE *output_file, GlassObject *object)
{
  GlassMaterial *mat_ptr = object->materials;

  while(mat_ptr != NULL)
  {
    _glassWriteMaterial(output_file, mat_ptr);
    fprintf(output_file, "\n");

    mat_ptr = mat_ptr->next;
  }

  return TRUE;
}

/* Write a material to the file */
int _glassWriteMaterial(FILE *output_file, GlassMaterial *material)
{
  /* Begin a material */
  fprintf(output_file, "begin material\n");

  fprintf(output_file, "  name\t\"%s\"\n", _glassMaterialGetName(material));
  if(material->texture)
    fprintf(output_file, "  texture\t\"%s\"\n",
	    _glassTextureGetName(_glassMaterialGetTexture(material)));
  fprintf(output_file, "\n");
  fprintf(output_file, "  ambient\t%f %f %f %f\n", material->ambient[0],
	  material->ambient[1], material->ambient[2], material->ambient[3]);
  fprintf(output_file, "  diffuse\t%f %f %f %f\n", material->diffuse[0],
	  material->diffuse[1], material->diffuse[2], material->diffuse[3]);
  fprintf(output_file, "  specular\t%f %f %f %f\n", material->specular[0],
	  material->specular[1], material->specular[2], material->specular[3]);
  fprintf(output_file, "  emission\t%f %f %f %f\n", material->emission[0],
	  material->emission[1], material->emission[2], material->emission[3]);
  fprintf(output_file, "  shininess\t%f\n", material->shininess);

  fprintf(output_file, "end\n");

  return TRUE;
}

/* Write all the triangle lists into the file */
int _glassWriteTriangleLists(FILE *output_file, GlassObject *object)
{
  GlassTriangleList *list_ptr = object->lists;

  while(list_ptr != NULL)
  {
    _glassWriteTriangleList(output_file, list_ptr);
    fprintf(output_file, "\n");

    list_ptr = list_ptr->next;
  }

  return TRUE;
}

/* Write a triangle list to the file */
int _glassWriteTriangleList(FILE *output_file, GlassTriangleList *tlist)
{
  /* Begin a triangle list */
  fprintf(output_file, "begin triangle_list\n");

  fprintf(output_file, "  name\t\"%s\"\n", _glassTriangleListGetName(tlist));

  _glassWriteTriangleListData(output_file, tlist);

  fprintf(output_file, "end\n");

  return TRUE;
}

/* Write the actual triangle data to the file */
int _glassWriteTriangleListData(FILE *output_file, GlassTriangleList *list)
{
  GlassTriangle *triangle_ptr = list->first_alpha;

  /* Write the alpha triangles */
  while(triangle_ptr != NULL)
  {
    _glassWriteTriangle(output_file, triangle_ptr);
    triangle_ptr = triangle_ptr->next;
  }

  /* Write the non-alpha triangles */
  triangle_ptr = list->first;
  while(triangle_ptr != NULL)
  {
    _glassWriteTriangle(output_file, triangle_ptr);
    triangle_ptr = triangle_ptr->next;
  }

  return TRUE;
}

/* Write one triangle to the file */
int _glassWriteTriangle(FILE *output_file, GlassTriangle *triangle)
{
  int i;

  fprintf(output_file, "\n");
  fprintf(output_file, "{");

  for(i = 0; i < 3; i++)
  {
    fprintf(output_file, "\n");
    if(triangle->material != NULL)
      fprintf(output_file, "  material\t\"%s\"\n\n", triangle->material->name);

    fprintf(output_file, "  pos\t%d %f %f %f\n",
	    i,
	    triangle->pos[i][0],
	    triangle->pos[i][1],
	    triangle->pos[i][2]);
    fprintf(output_file, "  col\t%d %f %f %f\n",
	    i,
	    triangle->col[i][0],
	    triangle->col[i][1],
	    triangle->col[i][2]);
    if(triangle->is_alpha)
      fprintf(output_file, "  alpha\t%d %f\n",
	      i,
	      triangle->col[i][3]);
    fprintf(output_file, "  nor\t%d %f %f %f\n",
	    i,
	    triangle->nor[i][0],
	    triangle->nor[i][1],
	    triangle->nor[i][2]);
    fprintf(output_file, "  tex\t%d %f %f\n",
	    i,
	    triangle->tex[i][0],
	    triangle->tex[i][1]);
  }

  fprintf(output_file, "}\n");

  return TRUE;
}

/* Write the variable to the file */
int _glassWriteVariables(FILE *output_file, GlassObject *object)
{
  GlassVariable *variable_ptr = object->variables;

  while(variable_ptr != NULL)
  {
    _glassWriteVariable(output_file, variable_ptr);
    fprintf(output_file, "\n");

    variable_ptr = variable_ptr->next;
  }

  return TRUE;
}

/* Write a variable to the file */
int _glassWriteVariable(FILE *output_file, GlassVariable *variable)
{
  /* Begin a variable */
  fprintf(output_file, "begin variable\n");

  /* Index/Min/Max/Value */
  fprintf(output_file, "  name\t\t\"%s\"\n", variable->name);
  fprintf(output_file, "  min\t\t%f\n", variable->min);
  fprintf(output_file, "  max\t\t%f\n", variable->max);
  fprintf(output_file, "  value\t\t%f\n", variable->value);

  fprintf(output_file, "end\n");

  return TRUE;
}

/* Write all the active points into the file */
int _glassWriteActivePoints(FILE *output_file, GlassObject *object)
{
  GlassActivePoint *apoint_ptr = object->points;

  while(apoint_ptr != NULL)
  {
    _glassWriteActivePoint(output_file, apoint_ptr);
    fprintf(output_file, "\n");

    apoint_ptr = apoint_ptr->next;
  }

  return TRUE;
}

/* Write an active point to the file */
int _glassWriteActivePoint(FILE *output_file, GlassActivePoint *apoint)
{
  /* Begin an active point */
  fprintf(output_file, "begin active_point\n");

  fprintf(output_file, "  name\t\t\"%s\"\n", _glassActivePointGetName(apoint));
  fprintf(output_file, "\n");
  fprintf(output_file, "  initial_dir\t%f %f %f\n", apoint->initial_dir[0],
	  apoint->initial_dir[1], apoint->initial_dir[2]);

  fprintf(output_file, "end\n");

  return TRUE;
}

/* Write the components and their component tree */
int _glassWriteComponents(FILE *output_file, GlassObject *object)
{
  GlassComponentList *clist = _glassComponentListNew(), *clist_ptr = clist;
  int index = 0, next_index, child_index;

  /* Make a component list, so indicies can be assigned */
  if(object->child != NULL)
    _glassMakeComponentList(clist, object->child, &index);
  else
    clist_ptr = NULL;

  /* Write the components */
  while(clist_ptr != NULL)
  {
    /* Get the indicies of the component, triangle list, next and child components,
       and the variables used */
    index = clist_ptr->index;
    next_index = _glassGetComponentIndex(clist, clist_ptr->component->next);
    child_index = _glassGetComponentIndex(clist, clist_ptr->component->child);

    /* Write the component */
    if(index != -1)
    {
      _glassWriteComponent(output_file, clist_ptr->component, index, next_index, child_index);
      fprintf(output_file, "\n");
    }

    clist_ptr = clist_ptr->next;
  }

  /* Get rid of the temporary component list */
  _glassFreeComponentList(clist);

  return TRUE;
}

/* Recursively make a component list, so indicies can be assigned to each component */
int _glassMakeComponentList(GlassComponentList *list,
			    GlassComponent *component, int *index)
{
  list->index = *index;
  list->component = component;

  if(component == NULL)
    return FALSE;

  /* Get the siblings */
  if(component->next != NULL)
  {
    list->next = _glassComponentListNew();
    (*index)++;
    _glassMakeComponentList(list->next, component->next, index);

    /* Skip past the components that were added */
    while(list->next != NULL)
      list = list->next;
  }

  /* Get the children */
  if(component->child != NULL)
  {
    list->next = _glassComponentListNew();
    (*index)++;
    _glassMakeComponentList(list->next, component->child, index);
  }

  return TRUE;
}

/* Get the index of a component */
int _glassGetComponentIndex(GlassComponentList *list, GlassComponent *component)
{
  while((list != NULL) && (list->component != component))
    list = list->next;

  if(list == NULL)
    return -1;
  else
    return list->index;
}

/* Write one component to the file */
int _glassWriteComponent(FILE *output_file, GlassComponent *component,
			 int index, int next_index, int child_index)
{
  /* Begin the component... */
  fprintf(output_file, "begin component\n");
  /* The indicies... */
  fprintf(output_file, "  index\t\t%d\n", index);
  if(next_index != -1)
    fprintf(output_file, "  next_index\t%d\n", next_index);
  if(child_index != -1)
    fprintf(output_file, "  child_index\t%d\n", child_index);
  /* The transform... */
  _glassWriteTransform(output_file, component->transform);
  /* The list name... */
  if(component->list)
    fprintf(output_file, "  list\t\t\"%s\"\n", component->list->name);
  /* The active point name... */
  if(component->apoint)
    fprintf(output_file, "  active_point\t\"%s\"\n", component->apoint->name);
  /* ...End the component */
  fprintf(output_file, "end\n");

  return TRUE;
}

/* Write a transform to the file */
int _glassWriteTransform(FILE *output_file, GlassTransform *transform)
{
  int i, nparams;

  if(transform == NULL)
    return TRUE;

  /* Write the transform */
  fprintf(output_file, "  transform\t%s", _glassGetTransformName(transform->type));
  nparams = _glassGetnParams(transform->type);
  for(i = 0; i < nparams; i++)
    if(transform->variables[i] == NULL) /* constant */
      fprintf(output_file, " %f", *transform->values[i]);
    else
      fprintf(output_file, " \"V::%s\"", transform->variables[i]->name);
  fprintf(output_file, "\n");

  return TRUE;
}

/* Convert a transform type to a string */
const char *_glassGetTransformName(GlassTransformType type)
{
  static char transforms[][20] = {"NO_TRANSFORM", "TRANSLATE", "ROTATE", "SCALE", "TRANSLATE_X",
				  "TRANSLATE_Y", "TRANSLATE_Z", "ROTATE_X", "ROTATE_Y",
				  "ROTATE_Z", "SCALE_X", "SCALE_Y", "SCALE_Z", "SCALE_EVEN"};

  switch(type)
  {
  case GLASS_NONE:
    return transforms[0];
    break;
  case GLASS_TRANSLATE:
    return transforms[1];
    break;
  case GLASS_ROTATE:
    return transforms[2];
    break;
  case GLASS_SCALE:
    return transforms[3];
    break;
  case GLASS_TRANSLATE_X:
    return transforms[4];
    break;
  case GLASS_TRANSLATE_Y:
    return transforms[5];
    break;
  case GLASS_TRANSLATE_Z:
    return transforms[6];
    break;
  case GLASS_ROTATE_X:
    return transforms[7];
    break;
  case GLASS_ROTATE_Y:
    return transforms[8];
    break;
  case GLASS_ROTATE_Z:
    return transforms[9];
    break;
  case GLASS_SCALE_X:
    return transforms[10];
    break;
  case GLASS_SCALE_Y:
    return transforms[11];
    break;
  case GLASS_SCALE_Z:
    return transforms[12];
    break;
  case GLASS_SCALE_EVEN:
    return transforms[13];
    break;
  default:
    fprintf(stderr, "Glass Warning: Unknown transform type (%d) (no name availiable)\n", type);
    return transforms[0];
    break;
  }
}
